<!-- table for file properties -->
{% macro m_table(lines, responsive=True) %}
  {%- if responsive %}
    <div class="table-responsive">
  {%- endif %}
<table class="table table-striped table-condensed" style="margin: 10px 0 15px 0;">
  <tbody>
  {%- for name, value in lines %}
    <tr class="row">
      <td class="col-md-3">{{ name }}</td>
      <td class="col-md-9">{{ value }}</td>
    </tr>
  {%- endfor %}
  </tbody>
</table>
{%- if responsive %}
  </div>
{%- endif %}
{%- endmacro %}
