{% from "macros/user.html" import m_user_photo, m_user_link %}
{% from "macros/form.html" import m_field %}

{% macro m_attachments(entity) %}
  {%- set display_attachments = attachments.manager(entity).macros['m_attachments'] %}
  {{ display_attachments(entity) }}
{% endmacro %}

{# single attachment #}
{% macro m_attachment(attachment) %}
  {%- set display_attachment = attachments.manager(entity).macros['m_attachment'] %}
  {{ display_attachment(attachment) }}
{% endmacro %}

{# new attachment form #}
{% macro m_attachment_form(entity) %}
  {%- set display_attachment_form = attachments.manager(entity).macros['m_attachment_form'] %}
  {{ display_attachment_form(entity) }}
{% endmacro %}
