
{% extends "base.html" %}

{% block main %}

<div class="row">

  <div class="col-sm-6 col-xs-12 col-sm-offset-3 login-form">
    <h2>{{ _("Forgot password?") }}</h2>

    <form action="{{ url_for(".forgotten_pw") }}" method="POST" name="login" role="form">
      {{ csrf.field() }}
      <div class="form-group">
        <input type="email" name="email" placeholder="E-Mail" autofocus="autofocus" class="form-control" />
      </div>
      <div class="form-group">
        <button class="btn btn-primary" type="submit">
          {{ _("Send me a new password") }}
        </button>
      </div>
    </form>
  </div>
</div>
{% endblock %}
